/*
 * Implementation of class ColorImage
 *
 * $Date: 2004/05/13 17:51:17 $
 * $Revision: 1.12 $
 *
 */

#include "WImage/StillImage.hh"
#include "WImage/ColorImage.hh"
#include "WImage/PgmReader.hh"
#include "WImage/PpmReader.hh"
#include "WImage/RawReader.hh"
#include "WImage/tools.h"
#ifdef PFI
#include "WImage/PfgReader.hh"
#include "WImage/PfcReader.hh"
#include "WImage/PfgWriter.hh"
#include "WImage/PfcWriter.hh"
#include <pfi.h>
#endif
#ifdef JPEG
#include "WImage/JpgReader.hh"
#include "WImage/JpgWriter.hh"
#endif
#include "WImage/PgmWriter.hh"
#include "WImage/PpmWriter.hh"
#include "WImage/RawWriter.hh"
#include "WImage/debug.h"
#include <stdexcept>
#include <iostream>
#include <assert.h>

#ifndef __WATCOMC__
using namespace std;
#endif

void 
ColorImage::init (int colors, clrmodel cmodel, Image **images)
{
  this->m_colors = colors;
  this->m_cmodel = cmodel;
  this->m_quality = 100;
  
  if (m_colors > 0)
    {
      m_images = new Image* [m_colors];
      
      for (int i = 0; i < m_colors; i++)
        {
          m_images[i] = images == NULL? new StillImage (0, 0)
            : images[i]->clone ();
        }
    }
  else
    {
      m_images = NULL;
    }
  m_xsize = m_ysize = m_xysize = 0;
}

void
ColorImage::destroy (void)
{
  if (m_images != NULL)
    {
      for (int i = 0; i < m_colors; i++)
        {
          if (m_images[i] != NULL)
            {
              delete m_images[i];
            }
        }
      delete m_images;
    }    
  this->m_cmodel = cm_rgb;
  m_xsize = m_ysize = m_xysize = 0;
  this->m_colors = 0;
}

ColorImage::ColorImage (int colors, clrmodel cmodel, Image **images) 
{
  init (colors, cmodel, images);
}

ColorImage::~ColorImage (void)
{
  destroy ();
}

void 
ColorImage::read (char const * fname)
{
  FileName fn (fname);
  read (fname, fn.guess ());
}

void
ColorImage::read (char const * fname, filetype ftype)
{
  GreymapReader *gr = NULL;
  PixmapReader *pr = NULL;
  ImageArray <coeff> *coeffs = coeffs = new ImageArray <coeff> (0, 0);
  
  switch (ftype)
    {
    case fn_ppm:
      destroy ();
      init (3);
      pr = new PpmReader (fname, m_images);
      break;
#ifdef JPEG
    case fn_jpg:
      destroy ();
      init (3);
      pr = new JpgReader (fname, m_images);
      break;
		std::cout << "fooo!!!";
#endif
    case fn_pgm:
      gr = new PgmReader (fname, *coeffs);
      break;
#ifdef PFI
    case fn_pfi:
      {
        FILE *tmp = fopen (fname, "r");
        int rows, cols, colors, colormodel;
        if (pfi_properties (tmp, &rows, &cols, &colors, &colormodel) < 0)
          {
            throw invalid_argument ("read: corrupt PFI file.");
          }
        if (colors == 1)
          {
            gr = new PfgReader (fname, *coeffs);
          }
        else 
          {
            destroy ();
            init (colors);
            pr = new PfcReader (fname, m_images, colors);
          }
      }
      break;
#endif
    default:
      throw invalid_argument ("read: unknown file type.");
      break;
    }

  if (gr != NULL)
    {
      /* this is pretty inefficient but not at all critical anyway */
      gr->read ();
      destroy ();
      m_colors = 1;
      m_images = new Image* [m_colors];
      m_images[0] = new StillImage (*coeffs);
      delete gr;
      m_cmodel = cm_grey;
    }
  else
    {
      pr->read ();
      if (pr->channels () == 1)
        {
          // memory already released in the reader
          for (int i = pr->channels (); i < m_colors; i++)
            {
              m_images[i] = NULL;
            }
          m_cmodel = cm_grey;
          m_colors = pr->channels ();
        }
      else
        {
          m_cmodel = pr->colormodel ();
        }
      delete pr;
    }
  delete coeffs;
  
  m_ysize = m_images[0]->rows ();
  m_xsize = m_images[0]->cols ();
  m_xysize = m_images[0]->size ();
  assert (m_cmodel != cm_unknown);
}

void 
ColorImage::write (char const *fname, bool beautify)
{
  if (m_colors < 1)
    {
      throw invalid_argument ("write: empty image.");
    }
  
  FileName fn (fname);
  write (fname, fn.guess (), beautify);
}

void
ColorImage::write (char const *fname, filetype ftype, bool beautify)
{
  GreymapWriter *gw = NULL;
  PixmapWriter *pw = NULL;
  bool ispixel = true;
  ImageArray <coeff> *coeffs = coeffs = new ImageArray <coeff> (0, 0);
  
  switch (ftype)
    {
    case fn_ppm:
      if (m_colors != 3)
        {
          throw invalid_argument ("write: number of colors is not 3");
        }
      pw = new PpmWriter (fname, m_images);
      break;
#ifdef JPEG
    case fn_jpg:
      pw = new JpgWriter (fname, m_images, m_colors, m_quality);
      break;
#endif
    case fn_pgm:
      if (m_colors > 1)
        {
          throw invalid_argument ("write: color reduction "
                                  "is source-format-dependent.");
        }
      gw = new PgmWriter (fname, *coeffs);
      break;
#ifdef PFI
    case fn_pfi:
      if (m_colors == 1)
        {
          gw = new PfgWriter (fname, *coeffs);
        }
      else
        {
          pw = new PfcWriter (fname, m_images, m_colors, m_cmodel);
        }
      ispixel = false;
      break;
#endif
    default:
      throw invalid_argument ("write: unknown file type.");
      break;
    }
  
  ColorImage *backup = NULL;
  if (beautify
    || (ispixel && (m_images[0]->anasteps () - m_images[0]->synsteps () > 0 
                    || (this->smin () < 0))))
    {
      backup = this->clone ();
      this->beautify ();
    }

  if (gw != NULL)
    {
      /* this is pretty inefficient but not at all critical anyway */
      coeffs->resize (m_images[0]->rows (), m_images[0]->cols ());
      for (int y = 0; y < m_images[0]->rows (); y++)
        {
          for (int x = 0; x < m_images[0]->cols (); x++)
            {
              coeffs->to (y, x, m_images[0]->at (y, x));
            }
        }
      gw->write ();
      delete gw;
    }
  else
    {
      pw->write ();
      delete pw;
    }

  if (backup != NULL)
    {
      for (int i = 0; i < m_colors; i++)
        {
          // We must not change the addresses here!
          m_images[i]->copy (backup->channel (i));
        }
      delete backup;
    }
  delete coeffs;
}

Image*
ColorImage::averageImage (void)
{
  Image *ret = m_images[0]->clone ();
  coeff avg;
  
  for (int i = 0; i < m_xysize; i++)
    {
      avg = 0;
      for (int j = 0; j < m_colors; j++)
        {
          avg += m_images[j]->at (i);
        }
      ret->to (i, avg / m_colors);
    }
  return ret;
}

coeff 
ColorImage::averageColor (int fromY, int fromX, int toY, int toX)
{
  coeff ret = 0;
  for (int i = 0; i < m_colors; i++)
    {
      ret += m_images[i]->saverage (fromY, fromX, toY, toX);
    }
  return ret / m_colors;
}

void 
ColorImage::colormodel (clrmodel model)
{
  if (m_colors != 3)
    {
      throw invalid_argument ("colormodel: number of colors is not 3.");
    }

  if (m_cmodel == model)
    {
      return;
    }

  /* This is from 
   * http://www.neuro.sfc.keio.ac.jp/~aly/polygon/info/color-space-faq.html
   */
  coeff rgb2yuv[] = {
    +0.299, +0.587, +0.114,
    -0.147, -0.289, +0.436,
    +0.615, -0.515, -0.100
  };
  coeff yuv2rgb[] = {
    +1.000, +0.000, +1.140,
    +1.000, -0.396, -0.581,
    +1.000, +2.029, +0.000
  };

  coeff *matrix;

  switch (model)
    {
    case cm_yuv:
      switch (m_cmodel)
        {
        case cm_rgb:
          matrix = rgb2yuv;
          break;
        default:
          throw invalid_argument ("colormodel: unsupported conversion.");
        }
      break;
    case cm_rgb:
      switch (m_cmodel)
        {
        case cm_yuv:
          matrix = yuv2rgb;
          break;
        default:
          throw invalid_argument ("colormodel: unsupported conversion.");
        }
      break;
    default:
      throw invalid_argument ("colormodel: unsupported format.");
    }
  
  coeff src[3];
  coeff dest;
  for (int i = 0; i < m_xysize; i++)
    {
      for (int j = 0; j < m_colors; j++)
        {
          src[j] = m_images[j]->at (i);
        }
      for (int j = 0; j < m_colors; j++)
        {
          dest = 0.0;
          for (int k = 0; k < m_colors; k++)
            {
              dest += matrix[j * m_colors + k] * src[k];
            }
          m_images[j]->to (i, dest);
        }
    }
  m_cmodel = model;
}

coeff
ColorImage::smin (void) const
{
  if (m_colors < 1)
    {
      throw invalid_argument ("smin: empty image.");
    }
  
  coeff result = m_images[0]->smin ();
  for (int i = 1; i < m_colors; i++)
    {
      coeff tmp = m_images[i]->smin ();
      if (tmp < result)
        {
          result = tmp;
        }
    }
  return result;
}

coeff
ColorImage::smax (void) const
{
  if (m_colors < 1)
    {
      throw invalid_argument ("smax: empty image.");
    }
  
  coeff result = m_images[0]->smax ();
  for (int i = 1; i < m_colors; i++)
    {
      coeff tmp = m_images[i]->smax ();
      if (tmp > result)
        {
          result = tmp;
        }
    }
  return result;
}

void 
ColorImage::beautify (int ysize, int xsize, int yoffs, int xoffs)
{
  if (ysize > 0 && xsize > 0)
    {
      Image **tempchannels = new Image* [m_colors];
      for (int i = 0; i < m_colors; i++)
        {
          tempchannels[i] = m_images[i]->crop (yoffs, xoffs, ysize, xsize);
        }
      ColorImage templl (m_colors, m_cmodel, tempchannels);
      
      for (int i = 0; i < m_colors; i++)
        {
          delete tempchannels[i];
        }
      
      delete tempchannels;

      coeff fill = (smax () - smin ()) / 2;
      
      for (int row = 0; row < ysize; row++)
        {
          for (int col = 0; col < xsize; col++)
            {
              for (int ch = 0; ch < m_colors; ch++)
                {
                  m_images[ch]->to (yoffs + col, xoffs + row, fill);
                }
            }
        }
      
      templl.valadjust ();
      valadjust ();

      for (int i = 0; i < m_colors; i++)
        {
          m_images[i]->paste (yoffs, xoffs, templl.channel (i));
        }
    }
  else 
    {
      valadjust ();
    }
}

void 
ColorImage::valadjust (void)
{
  coeff cmin = smin ();
  coeff cmax = smax ();
  coeff norm = 1.0;
  
  DPRINTF (("old max: %.2f, min: %.2f\n", cmax, cmin));
  
  if (cmax - cmin > 255.0)
    {
      norm = 255.0 / (cmax - cmin);
    }
  if (cmin < 0.0)
    {
      cmin *= -1;
      cmax = 0.0;
    }
  else if (cmax > 255.0)
    {
      cmax -= 255.0 / norm;
      cmin = 0.0;
    }
  else
    {
      cmax = cmin = 0;
    }
  
  DPRINTF (("new max: %.2f, min: %.2f\n", cmax, cmin));
  for (int i = 0; i < m_images[0]->size (); i++)
    {
      for (int j = 0; j < m_colors; j++)
        {
          coeff d = norm * (cmin + m_images[j]->at (i) - cmax);
          m_images[j]->to (i, d < 0? 0: d);
        }
    }
  assert (smin () >= 0.0);
  assert (smax () < 256);
  DPRINTF (("now max: %.2f, min: %.2f\n\n", smax (), smin ()));
}

ColorImage *
ColorImage::clone (void)
{
  return new ColorImage (m_colors, m_cmodel, m_images);
}

void 
ColorImage::truncate (coeff min, coeff max)
{
  for (int i = 0; i < m_xysize; i++)
    {
      for (int j = 0; j < m_colors; j++)
        {
          if (m_images[j]->at (i) < min)
            {
              m_images[j]->to (i, min);
            }
          else if (m_images[j]->at (i) > max)
            {
              m_images[j]->to (i, max);
            }
        }
    }
  assert (smin () >= 0.0);
  assert (smax () < 256);
}

bool
ColorImage::equals (ColorImage &img) const
{
  return this->epsilons (img, COEFF_EPSILON); 
}


bool
ColorImage::epsilons (ColorImage &img, coeff epsilon) const
{
  for (int i = 0; i < m_colors; i++)
    {
      if (!m_images[i]->epsilons (img.channel (i), epsilon))
        {
          return false;
        }
    }
  return true;
}

unsigned char*
ColorImage::toPixmap (clrmodel cmodel)
{
  clrmodel old = m_cmodel;
  if (old !=  cmodel)
    {
      colormodel (cmodel);
    }

  
  unsigned char *allpixels = new unsigned char [m_colors * m_xysize];

  /* import */
  for (int i = 0; i < m_xysize; i++)
    {
      for (int j = 0; j < m_colors; j++)
        {
          allpixels[m_colors * i + j] 
            = (unsigned char)tools_coeff2pixel (m_images[j]->at (i));
        }
    }
  
  if (old !=  cmodel)
    {
      colormodel (old);
    }

  return allpixels;
}

