/*
 * Implementation of class Image
 *
 * $Date: 2004/05/21 09:22:58 $
 * $Revision: 1.9 $
 *
 */

#include "WImage/Image.hh"
#include "WImage/debug.h"
#include "WImage/miscdefs.h"
#include "WImage/tools.h"
#include <assert.h>
#include <math.h>
#include <stdexcept>
#include <iostream>

#ifndef __WATCOMC__
using namespace std;
#endif


Image::Image (void)
{ 
  m_xsize = 0; 
  m_ysize = 0; 
  m_xysize = 0; 
  m_acol = NULL; 
  m_arow = NULL; 
  m_afull = NULL; 
}

void 
Image::valadjust (void)
{
  coeff cmin = smin ();
  coeff cmax = smax ();
  coeff norm = 1.0;
  
  DPRINTF (("old max: %.2f, min: %.2f\n", cmax, cmin));
  
  if (cmax - cmin > 255.0)
    {
      norm = 255.0 / (cmax - cmin);
    }
  if (cmin < 0.0)
    {
      cmin *= -1;
      cmax = 0.0;
    }
  else if (cmax > 255.0)
    {
      cmax -= 255.0 / norm;
      cmin = 0.0;
    }
  else
    {
      cmax = cmin = 0;
    }
  
  DPRINTF (("new max: %.2f, min: %.2f\n", cmax, cmin));
  for (int i = 0; i < size (); i++)
    {
      coeff d = norm * (cmin + at (i) - cmax);
      to (i, d < 0? 0: d);
    }
  assert (smin () >= 0.0);
  assert (smax () < 256);
  DPRINTF (("now max: %.2f, min: %.2f\n\n", smax (), smin ()));
}

void 
Image::dimcheck (const Image &img)
{
  if (this->rows () != img.rows () || this->cols () != img.cols ())
    {
      throw invalid_argument ("Different dimensions.");
    }
}

Image *
Image::crop (int yoffs, int xoffs, int ysize, int xsize)
{
  Image *ret = mkImage
    ((yoffs + ysize) > this->m_ysize? this->m_ysize - yoffs: ysize,
     (xoffs + xsize) > this->m_xsize? this->m_xsize - xoffs: xsize);
  
  for (int row = 0; row < ysize; row++)
    {
      for (int col = 0; col < xsize; col++)
        {
          int y = yoffs+row, x = xoffs+col;
          if (y < this->m_ysize && x < this->m_xsize)
            {
              ret->to (row, col, this->at (y, x));
            }
        }
    }
  return ret;
}

void 
Image::paste (int yoffs, int xoffs, Image &img)
{
  for (int row = 0; row < img.rows (); row++)
    {
      for (int col = 0; col < img.cols (); col++)
        {
          if (row < this->m_ysize && col < this->m_xsize)
            {
              this->to (yoffs+row, xoffs+col, img.at (row, col));
            }
        }
    }
}

void
Image::copy (Image &img)
{
  resize (img.rows (), img.cols ());
  for (int row = 0; row < img.rows (); row++)
    {
      for (int col = 0; col < img.cols (); col++)
        {
          this->to (row, col, img.at (row, col));
        }
    }
}

coeff 
Image::sqvariance (int fromY, int fromX, int toY, int toX) const
{
  int nVals = 0;
  int lower = (toY < 0)? m_ysize: toY;
  int right = (toX < 0)? m_xsize: toX;
  coeff avg = this->saverage (fromY, fromX, lower, right);
  coeff ret = 0.0;

  for (int y = (fromY < 0? 0: fromY); y < lower && y < m_ysize; y++)
    {
      for (int x = (fromX < 0? 0: fromX); x < right && x < m_xsize; x++)
        {
          ++nVals;
          ret += this->at (y, x) * this->at (y, x);
        }
    }
  
  return (ret / nVals) - avg * avg;
}

coeff
Image::sdeviation (int fromY, int fromX, int toY, int toX) const
{
  int nVals = 0;
  int lower = (toY < 0)? m_ysize: toY;
  int right = (toX < 0)? m_xsize: toX;
  coeff avg = this->saverage (fromY, fromX, lower, right);
  coeff ret = 0.0;
  coeff tmp = 0.0;

  for (int y = (fromY < 0? 0: fromY); y < lower && y < m_ysize; y++)
    {
      for (int x = (fromX < 0? 0: fromX); x < right && x < m_xsize; x++)
        {
          ++nVals;
          tmp = this->at (y, x) - avg;
          ret += tmp * tmp;
        }
    }
  
  return sqrt (ret / (nVals - 1));
}

double 
Image::iproduct (Image &img)
{
  coeff rc = 0.0;
  if (m_ysize != img.rows () || m_xsize != img.cols ())
    {
      throw invalid_argument ("iproduct: the image "
                              "dimensions are not identical.");
    }
  for (int i = 0; i < m_xysize; i++)
    {
      rc += this->at (i) * img.at (i);
    }
  
  return (double)rc;
}

void 
Image::normalize (coeff factor)
{ 
  for (int i = 0; i < this->size (); i++)
    {
      this->to (i, this->at (i) / factor);
    }
}

void 
Image::unnormalize (coeff factor)
{ 
  for (int i = 0; i < this->size (); i++)
    {
      this->to (i, this->at (i) * factor);
    }
}

void 
Image::beautify (int ysize, int xsize, int yoffs, int xoffs)
{
  if (ysize > 0 && xsize > 0)
    {
      Image *templl = crop (yoffs, xoffs, ysize, xsize);
      coeff fill = (smax () - smin ()) / 2;
      
      for (int row = 0; row < ysize; row++)
        {
          for (int col = 0; col < xsize; col++)
            {
              assert (yoffs + row < m_ysize);
              assert (xoffs + col < m_xsize);
              to (yoffs + row, xoffs + col, fill);
            }
        }
      
      templl->valadjust ();
      valadjust ();
      paste (yoffs, xoffs, *templl);
      delete templl;
    }
  else 
    {
      valadjust ();
    }
}

void 
Image::truncate (coeff min, coeff max)
{
  for (int i = 0; i < this->size (); i++)
    {
      if (this->at (i) < min)
        {
          this->to (i, min);
        }
      else if (this->at (i) > max)
        {
          this->to (i, max);
        }
    }
  assert (smin () >= 0);
}
  
void 
Image::shift (int yoffs, int xoffs)
{
  Image *temp1=0;
  Image *temp2=0;
  
  if (xoffs < 0)
    {
      temp1=this->crop (0, -xoffs,this->m_ysize,this->m_xsize+xoffs);
      temp2=this->crop (0, 0,this->m_ysize,-xoffs);
      paste (0, 0, *temp1);
      paste (0, this->m_xsize+xoffs, *temp2);
    }
  else if (xoffs > 0)
    {
      temp1=this->crop (0, this->m_xsize-xoffs,this->m_ysize,xoffs);
      temp2=this->crop (0, 0,this->m_ysize,this->m_xsize-xoffs);
      paste (0, 0, *temp1);
      paste (0, xoffs, *temp2);
    }
  delete temp1;
  delete temp2;
  temp1=temp2=0;

  if (yoffs < 0)
    {
      temp1=this->crop (-yoffs,0,this->m_ysize+yoffs,this->m_xsize);
      temp2=this->crop (0, 0,-yoffs,this->m_xsize);
      paste (0, 0, *temp1);
      paste (this->m_ysize+yoffs,0, *temp2);
    }
  else if (yoffs > 0)
    {
      temp1=this->crop (this->m_ysize-yoffs,0,yoffs,this->m_xsize);
      temp2=this->crop (0, 0,this->m_ysize-yoffs,this->m_xsize);
      paste (0, 0, *temp1);
      paste (yoffs,0, *temp2);
    }
  delete temp1;
  delete temp2;
}


void 
Image::importPixels (int rows, int cols, pixel *buf)
{
  resize (rows, cols);
  
  anasteps (0);
  synsteps (0);

  for (int i = 0; i < m_xysize; i++)
    {
      this->to (i, (coeff)buf[i]);
    }
}

pixel * 
Image::exportPixels (void)
{
  pixel *ret = new pixel [m_xysize];
  for (int i = 0; i < m_xysize; i++)
    {
      ret[i] = tools_coeff2pixel (this->at (i));
    }
  return ret;
}

void 
Image::importCoeffs (int rows, int cols, coeff *buf)
{
  resize (rows, cols);
  
  anasteps (0);
  synsteps (0);

  for (int i = 0; i < m_xysize; i++)
    {
      this->to (i, buf[i]);
    }
}

coeff * 
Image::exportCoeffs (void)
{
  coeff *ret = new coeff [m_xysize];
  for (int i = 0; i < m_xysize; i++)
    {
      ret[i] = this->at (i);
    }
  return ret;
}
