/*
 * Implementation of class ImageComparison
 *
 * $Date: 2003/10/17 14:33:39 $
 * $Revision: 1.3 $
 *
 */

#include "Wave/StandardTransform.hh"
#include "WTools/ImageComparison.hh"
#include "WTools/ImageInformation.hh"
#include "WImage/miscdefs.h"
#include <math.h>
#include <assert.h>

#ifndef __WATCOMC__
using namespace std;
#endif

static void truncate (Image*, double);

ImageComparison::ImageComparison (const Image &img1, const Image &img2)
{
  m_images[0] = img1.clone ();
  m_images[1] = img2.clone ();
  m_ysize = MIN (m_images[0]->rows (), m_images[1]->rows ());
  m_xsize = MIN (m_images[0]->cols (), m_images[1]->cols ());
  m_xysize = m_xsize * m_ysize;
  
  for (int i = 0; i < 2; i++)
    {
      if (m_images[i]->rows () != m_ysize || m_images[i]->cols () != m_xsize)
        {
          m_images[i]->resize (m_ysize, m_xsize);
        }
    }
  
  m_lqcachesize = MAX(m_ysize, m_xsize);
  m_lqcache = new logvals [m_lqcachesize];

  for (int i = 0; i < m_lqcachesize; i++)
    {
      m_lqcache[i].init = false;
    }

  double defaults[][6] = { { 4.04, 0.78, 0.46, 0.42, 0.41, 0.32 }, 
                           { 5.00, 0.83, 1.01, 0.52, 0.47, 0.30 }};

  for (int row = 0; row < 2; row++)
    {
      for (int col = 0; col < 6; col++)
        {
          m_weights[row][col] = defaults[row][col];
        }
    }
}

ImageComparison::~ImageComparison (void)
{
  delete m_images[0];
  delete m_images[1];
  delete [] m_lqcache;
}

Image*
ImageComparison::difference (coeff magnify, bool reverse)
{
  Image *ret = m_images[0]->clone ();
  coeff tmp;
  
  for (int row = 0; row < m_ysize; row++)
    {
      for (int col = 0; col < m_xsize; col++)
        {
          tmp = (reverse? -1: 1) * ret->at (row, col) 
            + (reverse? 1: -1) * m_images[1]->at (row, col);
          ret->to (row, col, magnify*tmp);
        }
    }
  return ret;
}


double 
ImageComparison::cmpsnr (void)
{
  double total_error = 0.0;
  
  for (int i = 0; i < m_xysize; i++)
    {
      total_error += SQUARE (m_images[0]->at (i) - m_images[1]->at (i));
    }

  double rms = sqrt(total_error/(double)(m_xysize));
  double psnr = 20.0 * log(255.0/rms)/log(10.0);

  return psnr;
}

double 
ImageComparison::cmpmse (void)
{
  double total_error = 0.0;
  
  for (int i = 0; i < m_xysize; i++)
    {
      total_error += SQUARE (m_images[0]->at (i) - m_images[1]->at (i));
    }

  double mse = total_error/(double)(m_xysize);
  
  return mse;
}

lq
ImageComparison::distlq (double percent, FilterSet &flt, imgtype type)
{
  double sum1, sum2;
  lq score = { 0.0, 0.0 };

  WaveletTransform *transforms[2];
  Image *origs[2];

  for (int i = 0; i < 2; i++)
    {
      origs[i] = m_images[i]->clone ();
      transforms[i] = new StandardTransform (*m_images[i], flt);
      transforms[i]->analysis (0);
      delete transforms[i];
    }

  score.colors = m_weights[type][0] 
    * fabs (m_images[0]->at (0, 0) - m_images[1]->at (0, 0));

  truncate (m_images[0], percent);
  truncate (m_images[1], percent);

  sum1 = lq_sum (type, true);
  sum2 = lq_sum (type, false);
  
  /* create a normalized details measure */
  score.details = (sum1 < sum2? sum1: sum2) / (m_xysize * percent / 10000.0);

  for (int i = 0; i < 2; i++)
    {
      delete m_images[i];
      m_images[i] = origs[i];
    }
  
  return score;
}

int 
ImageComparison::level (int i)
{
  if (!m_lqcache[i].init)
    {
      m_lqcache[i].init = true;
      m_lqcache[i].val = (int)floor (log ((double)i) / log ((double)2));
    }
  return m_lqcache[i].val;
}

double 
ImageComparison::weight (int y, int x, imgtype type)
{
  assert (y > 0);
  assert (x > 0);
  int l1 = level (y);
  int l2 = y == x? l1: level (x);
  return m_weights [type][MIN (MAX (l1, l2), 5)];
}

double 
ImageComparison::lq_sum (imgtype type, bool reverse)
{
  double score = 0;
  
  for (int row = 1; row < m_ysize; row++)
    {
      for (int col = 1; col < m_xsize; col++)
        {
          coeff c1 = m_images[0]->at (row, col);
          if (fabs (c1) - COEFF_EPSILON > 0 && c1 != m_images[1]->at (row, col))
            {
              score += weight (row, col, type);
            }
        }
    }
  return score;
}

static void 
truncate (Image *img, double percent)
{
  CoeffInformation *ci;
  ImageInformation *ii = new ImageInformation (*img);

  ii->asort ();
  ii->shrink ((int)((double)(img->size () * percent) / 100.0), false);

  for (int i = 0; i < img->size (); i++)
    {
      img->to (i, 0.0);
    }
  
  for (int i = 0; i < ii->size (); i++)
    {
      ci =  &ii->at (i);
      img->to (ci->ypos (), ci->xpos (), ci->val () > 0? 1: -1);
    }
  delete ii;
}
