/*
 * class ColorImage
 *
 * $Date: 2003/09/09 17:29:46 $
 * $Revision: 1.10 $
 *
 */

#ifndef COLOR_IMAGE_HH__
#define COLOR_IMAGE_HH__

#include "WImage/Image.hh"

/** @addtogroup WImage */

/**
 * @class ColorImage
 * An image wrapper for bitmap images of three color components.
 * Objects may import and export different bitmap formats,
 * such as PPM. The color models can be changed (RGB or YUV).
 */
class 
ColorImage
{
public:
  /** Constructor. Creates objects and sets actual start values.
   * @param colors
   * the number of colors, and therefore the array's length 
   * @param cmodel 
   * the color model
   * @param images
   * an array of {\em colors} images to copy the channel data from */
  ColorImage (int colors = 3, clrmodel cmodel = cm_rgb, Image **images = NULL);
  /** Destructor. Frees allocated objects. */
  virtual ~ColorImage (void);

  /** Returns a reference to one of the color channels.
   * {\em Don't change the dimensions of single channels unless you know
   * what you're doing!}
   * @param num the number of the color channel
   * @return the color channel as a reference to an Image object */
  inline Image &channel (int num) { return *m_images[num]; }
  
  /** Read the image. All steps independent of the file format will be 
   * performed, like testing for file readability etc. 
   * @exception invalid_argument
   * invalid file format
   * @exception ios_base::failure
   * a read error has occured [not supported by all libraries, so eventually 
   * {\em invalid_argument} instead] */
  void read (char const *fname);
  /** Read an image. Reads an image from a file using the specified 
   * file type. Currently PGM, RAW and PFI are supported.
   * @exception invalid_argument
   * the file type is either not supported or could not be determined
   * from the given file name
   * @param fname
   * the file name, if {\em NULL}, then {\em stdin}
   * @param ftype
   * the file type */
  void read (char const *fname, filetype ftype);

  /** Write the image. All steps independent of the file format will be 
   * performed, like testing for file writeability etc. 
   * @exception ios_base::failure
   * a read error has occured [not supported by all libraries, so eventually 
   * {\em invalid_argument} instead] 
   * @exception invalid_argument
   * the source image has a different number of colors than the target 
   * format allows
   * @param fname
   * the target file name
   * @param beautify
   * beautify images that have not had more analysis than synthesis steps? */
  void write (char const *fname, bool beautify = false);
  /** Write an image (abstract). Writes an image to a file using the 
   * specified file type.
   * @param ftype
   * the file type
   * @param fname
   * the file name, if {\em NULL}, then {\em stdout} 
   * @param beautify
   * beautify images that have not had more analysis than synthesis steps? */
  void write (char const *fname, filetype ftype, bool beautify = false);

  /** Return the current colormodel
   * @exception invalid_argument
   * number of colors is not 3, or unsupported format / conversion
   * @return the colormodel */
  inline clrmodel colormodel (void) { return m_cmodel; }
  /** Return the current number of colors
   * @return the number of colors */
  inline int colors (void) { return m_colors; }
  /** Convert the image to a new colormodel. Note: this operation
   * is not completely invertable. Thus it should only be applied
   * if no further transforms and no back-conversion are desired.
   * @param model
   * the new colormodel ({\em cm_rgb} or {\em cm_yuv}) */
  void colormodel (clrmodel model);
  
  /** Return a new greyscale image containing the three color channels'
   * averages
   * @return the new image */
  Image *averageImage (void);

  /** Return the average color for a rectangular region inside the image
   * drawn from one point within and the second point just outside the
   * region
   * @param fromY the first point's row
   * @param fromX the first point's col
   * @param toY the second point's row (-1 if lower image border)
   * @param toX the second point's col (-1 if right image border)
   * @return the average color */
  coeff averageColor (int fromY = 0, int fromX = 0, 
                      int toY = -1, int toX = -1);

  /** Returns the maximum value. Signs will be considered. 
   * @exception invalid_argument
   * empty image
   * @return
   * the maximum */
  coeff smax (void) const;
  /** Returns the maximum value. Signs will be considered. 
   * @exception invalid_argument
   * empty image
   * @return
   * the maximum */
  coeff smin (void) const;

  /** Compares two images. Return {\em true} if both are equal.
   * @param img
   * The other {\em ColorImage} object
   * @return 
   * if equals: {\em true}, else {\em false} */
  bool equals (ColorImage &img) const;
  /** Rough comparison. See if two images are similar according to a given
   * {\em epsilon} (important for floating-point comparisons).
   * @param img
   * the other {\em ColorImage} object
   * @param epsilon
   * the epsilon
   * @return
   * if both are identical: {\em true}, else {\em false} */
  bool epsilons (ColorImage &img, coeff epsilon) const;

  /** Do a dumb kind of beautification by just truncating values to 
   * their possible maxima depending on the image format or whatever.
   * The original data will be overwritten, so this should only be 
   * used on clones or if the original data is not needed anymore.
   * @param min
   * the minimum allowed value
   * @param max
   * the maximum allowed value */
  void truncate (coeff min = 0, coeff max = 255);
  
  /** Do an intelligent kind of beautification considering previous
   * transform steps. The original data will be overwritten, so this
   * should only be used on clones or if the image is not going to be
   * transformed afterwise. One region inside the image could be the 
   * low-pass subband and should thus be handled separately. If that is 
   * desired, the {\em ysize} and {\em xsize} parameters should be set
   * to nonzero values. 
   * @param ysize
   * the horizontal size of the region that should be handled separately
   * @param xsize 
   * the vertical size of the region that should be handled separately
   * @param yoffs
   * the horizontal offset of the region that should be handled separately
   * @param xoffs 
   * the vertical offset of the region that should be handled separately */
  void beautify (int ysize = 0, int xsize = 0, int yoffs = 0, int xoffs = 0);

  /** Produce a copy of this image.
   * @return the copy */
  ColorImage *clone (void);

  /** Export the image for other applications. This only works if the image
   * has three channels!
   * @exception invalid_argument if the current colormodel or the one 
   * passed as parameter is unsupported.
   * @param cmodel the color model in which the image is exported. Currently
   * {\em cm_rgb} and {\em cm_yuv} are supported.
   * @return a new buffer containing the image pixels in the format according
   * on the cmodel parameter */
  unsigned char *toPixmap (clrmodel cmodel);

  /** Return the current image quality factor.
   * @return the current image quality factor */
  inline int quality (void) { return m_quality; }

  /** Sets a new image quality factor.
   * @return the new image quality factor */
  inline void quality (int quality) { m_quality = quality; }

  /** Return the number of rows.
   * @return
   * the number of rows */
  inline int rows (void) const { return m_images[0]->rows (); }
  /** Return the number of columns.
   * @return
   * the number of columns */
  inline int cols (void) const { return m_images[0]->cols (); }
  /** Return the overall size.
   * @return
   * the overall size */
  inline int size (void) const { return m_images[0]->size (); }

private:
  /* The way the current image's colors are encoded. */
  clrmodel m_cmodel;
  /** The number of pixel cols */
  int m_xsize;
  /** The number of pixel rows */
  int m_ysize;
  /** The overall number of pixels */
  int m_xysize;
  /** The number of colors */
  int m_colors;
  /** The image quality (if stored in lossy format like JPG) */
  int m_quality;
  /** An array of {\em colors} Image objects */
  Image **m_images;
protected:
  /** Allocate objects and set standard values.
   * @param colors
   * the number of colors 
   * @param cmodel
   * the initial color model
   * @param images
   * an array of {\em colors} images to copy the channel data from */
  void init (int colors, clrmodel cmodel = cm_rgb, Image **images = NULL);
  /** Deletes allocated objects. */
  void destroy (void);
  /** Adjust the coefficients to make them pixelizable. */
  void valadjust (void);
}; /* class ColorImage */

#endif /* COLOR_IMAGE_HH__ */

