/*
 * class Image
 *
 * $Date: 2003/11/14 14:57:41 $
 * $Revision: 1.6 $
 *
 */

#ifndef IMAGE_COMPARISON_HH__
#define IMAGE_COMPARISON_HH__

#include "WImage/Image.hh"
#include "Wave/Filter.hh"

/** @addtogroup WTools*/

/** The type of the image. Either hand-drawn or scanned. */
typedef enum { DRAWN = 0, SCANNED } imgtype;


/** A data structure for an $L^q$ norm. */
typedef struct lq {
  /** The weighted distance between the average colors */
  double colors;
  /** The weighted distances between the top coefficients */
  double details;
} lq;

/** A data structure for a cache of (log (i) / log (2)) entries. */
typedef struct logvals {
  int val;
  bool init;
} logvals;

/**
 * @class ImageComparison
 * A class for differences between two images 
 */
class ImageComparison 
{
public:
  /** Constructor. Creates copies of the two images. If the two images
   * have different dimensions they will be resized to the largest common
   * dimensions.
   * @param img1
   * the first image
   * @param img2
   * the second image */
  ImageComparison (const Image &img1, const Image &img2);
  /** Destructor. Releases allocated memory. */
  ~ImageComparison (void);

  /** Return the number of rows.
   * @return
   * the number of rows */
  inline int rows (void) const { return m_ysize; }
  /** Return the number of columns.
   * @return
   * the number of columns */
  inline int cols (void) const { return m_xsize; }
  /** Return the overall size.
   * @return
   * the overall size */
  inline int size (void) const { return m_xysize; }

  /**
   * Produce an image holding the difference between the first and the 
   * second image
   * @param magnify
   * multiply the result by a factor to increase / decrease visibility
   * @param reverse
   * reverse the subtraction (take img1 - img2)
   * @return
   * a new image containing the difference */
  Image *difference (coeff magnify = 1, bool reverse = false);

  /** Compare with another image producing the PSNR. Stolen from Geoff 
   * Davis' Wavelet Coder kit. 
   * @return
   * The PSNR */
  double cmpsnr (void);
  
  /** Compare with another image producing the mean square error. Stolen 
   * from Geoff Davis' Wavelet Coder kit.
   * @return
   * The MSE */
  double cmpmse (void);
  
  /** Calculate the distance to another image according to the $L^q$ norm.
   * Idea and basic algorithm from Jacobs, Finkelstein, Salesin: "Fast
   * multiresolution image querying", Proc. of SIGGRAPH, 95.
   * @exception invalid_argument
   * image is not square
   * @param percent
   * the relative number of Wavelet coefficients to compare
   * @param flt
   * the filter to use
   * @param type
   * the type of the images to be assumed (drawn or scanned)
   * @return
   * The $L^q$ norm */
  lq distlq (double percent, FilterSet &flt, imgtype type);

  /** Return the details component from $L^q$ distance between this 
   * and another image.
   * @exception invalid_argument
   * image is not square
   * @param percent
   * the relative number of Wavelet coefficients to compare
   * @param flt
   * the filter to use
   * @param type
   * the type of the images to be assumed (drawn or scanned)
   * @return
   * The $L^q_d$ norm */
  inline double distlqd (double percent, FilterSet &flt, imgtype type) { 
    return distlq (percent, flt, type).details; }
  
protected:
  /** Number of rows. */
  int m_ysize;
  /** Number of cols. */
  int m_xsize;
  /** Overall size. */
  int m_xysize;
  
  /** An array holding the two images to be compared. */
  Image *m_images[2];
  
  /** A cache needed by distlq() to avoid repeated calculation of 
   * floor (log (i) / log (2)). */
  logvals * m_lqcache;

  /** The six weights for scanned images, needed by distlq(), 
   * Y channel, according to Jacobs et.al. */ 
  double m_weights[2][6];

  /** The size of the m_lqcache. Always as big as MAX (rows, cols). */
  int m_lqcachesize;

  /** Helper method for distlq(). Calculates the weighing level
   * for a position
   * @param pos
   * the position (row or col)
   * @return 
   * the level */
  int level (int pos);

  /** Helper method for distlq(). Calculate the weight for a position (y, x). 
   * @param y
   * the row
   * @param x
   * the col
   * @param type
   * the images' type to be assumed (drawn or scanned)
   * @return 
   * the position's weight */
  double weight (int y, int x, imgtype type);
  
  /** Helper method for distlq(). Calculate the sum of differences of
   * the two decomposed images.
   * @param type
   * the type (drawn or scanned) to be assumed
   * @param reverse
   * if true, subract the first from the second image */
  double lq_sum (imgtype type, bool reverse);
};

#endif // IMAGE_COMPARISON_HH__
