/*
 * Implementation of class WaveletTransform
 *
 * $Date: 2004/05/19 14:16:27 $
 * $Revision: 1.13 $
 *
 */

//#undef DPRINTF
#include "Wave/WaveletTransform.hh"
#include "WImage/debug.h"
#include "WImage/tools.h"
#include <stdexcept>
#include <iostream>
#include <assert.h>
#include <math.h>

#ifndef __WATCOMC__
using namespace std;
#endif

static const area areas [4] = { LL, HL, LH, HH };

void
WaveletTransform::sanity (void)
{
  if (m_image->rows () != m_buffer->rows () 
      || m_image->cols () != m_buffer->cols ())
    {
      m_buffer->resize (m_image->rows (), m_image->cols ());
    }
}

void 
WaveletTransform::expandImage (void)
{
  int powerY = 0;
  int powerX = 0;
  
  if (!tools_powerOfTwo (m_image->rows (), powerY) ||
      m_image->rows () != m_image->cols ())
    {
      (void)tools_powerOfTwo (m_image->cols (), powerX);
      int newY, newX;
      m_rows = m_image->rows ();
      m_cols = m_image->cols ();
      newY = newX = TWOPOW (powerY > powerX? powerY: powerX);
      assert (newY >= m_rows);
      assert (newX >= m_cols);
      m_image->resize (newY, newX);
      sanity ();
    }
}


void 
WaveletTransform::restoreImage (void)
{
  m_image->resize (m_rows, m_cols);
  sanity ();
}


void 
WaveletTransform::analysis (int steps)
{
  //int power = 0;
  if (m_image->rows () != m_image->cols ())
    //|| !tools_powerOfTwo (m_image->rows (), power))
   // {
   //   throw invalid_argument ("analysis:"
   //                           " non square images are not supported.");
   // }
  sanity ();
  doanalysis (steps);
}

void 
WaveletTransform::synthesis (int steps, int prevSteps)
{
  //if (m_image->rows () != m_image->cols ())
  //  {
  //    throw invalid_argument ("synthesis:"
  //                            " non square images are not supported.");
  //  }
  //int power = 0;
  //if (!tools_powerOfTwo (m_image->rows (), power))
  //  {
  //    throw invalid_argument ("synthesis:"
  //                            " sidelengths need to be powers of two.");
  //  }
  sanity ();
  dosynthesis (steps, prevSteps);
}


WaveletTransform::WaveletTransform (Image &img, FilterSet &fil)
{ 
  m_image = &img; 
  m_filter = &fil;
  m_wavelet = new Wavelet (fil); 
  m_buffer = m_image->clone ();
  m_rows = m_image->rows ();
  m_cols = m_image->cols ();
}

WaveletTransform::~WaveletTransform (void)
{
  delete m_wavelet;
  delete m_buffer;
}


Image *
WaveletTransform::subband (area what, int steps)
{
  int yoffs = 0, xoffs = 0, ysize = 0, xsize = 0;
  where (what, steps, yoffs, xoffs, ysize, xsize);
  
  DPRINTF (("crop (%d, %d, %d, %d)\n", yoffs, xoffs, ysize, xsize));
  return m_image->crop (yoffs, xoffs, ysize, xsize);
}
  
void 
WaveletTransform::import (Image &img, area what, int steps)
{
  int yoffs = 0, xoffs = 0, ysize = 0, xsize = 0;
  where (what, steps, yoffs, xoffs, ysize, xsize);
  if (ysize != img.rows () || xsize != img.cols ())
    {
      throw invalid_argument ("import: the image "
                              "dimensions don't match the subband size.");
    }
  DPRINTF (("import: area: %d, yoffs: %d, xoffs: %d\n", 
            (int)what, yoffs, xoffs));
  m_image->paste (yoffs, xoffs, img);
}

void 
WaveletTransform::fill (coeff value, area what, int steps)
{
  int yoffs = 0, xoffs = 0, ysize = 0, xsize = 0;
  where (what, steps, yoffs, xoffs, ysize, xsize);
  DPRINTF (("fill: area: %d, yoffs: %d, xoffs: %d, value: %f\n", 
            (int)what, yoffs, xoffs, value));
  for (int row = 0; row < ysize; row++)
    {
      for (int col = 0; col < xsize; col++)
        {
          m_image->to (yoffs + row, xoffs + col, value);
        }
    }
}

void
WaveletTransform::where (area what, int subband, 
                         int &yoffs, int &xoffs, int &ysize, int &xsize)
{
  DPRINTF (("rows: %d, cols: %d\n", m_image->rows (), m_image->cols ()));
  if (subband < 0)
    {
      subband = this->steps ();
    }

  xoffs = yoffs = 0;
  
  DPRINTF (("subband: %d, anasteps: %d, synsteps: %d\n", 
            subband, m_image->anasteps (), m_image->synsteps ()));
  
  ysize = m_image->rows () / TWOPOW (subband);
  xsize = m_image->cols () / TWOPOW (subband);

  if (what == LL)
    {
      ysize = (m_image->rows () + 1)/ TWOPOW (subband);
      xsize = (m_image->cols () + 1)/ TWOPOW (subband);
    }
  else if (what == HH)
    {
      yoffs = (m_image->rows () + 1) / TWOPOW (subband);
      xoffs = (m_image->cols () + 1) / TWOPOW (subband);
    }
  else if (what == HL)
    {
      ysize = (m_image->rows () + 1)/ TWOPOW (subband);
      xoffs = (m_image->cols () + 1) / TWOPOW (subband);
    }
  else if (what == LH)
    {
      yoffs = (m_image->rows () + 1) / TWOPOW (subband);
      xsize = m_image->cols () / TWOPOW (subband);
    }
  
  DPRINTF (("area: %d, yoffs: %d, xoffs: %d, ysize: %d, xsize: %d\n", 
            (int)what, yoffs, xoffs, ysize, xsize));
  assert (yoffs + ysize <= m_image->rows ());
  assert (xoffs + xsize <= m_image->cols ());
}

void 
WaveletTransform::where (int ypos, int xpos, int &subband, area &channel, 
                         int steps)
{
  if (steps < 0)
    {
      steps = this->steps ();
    }
  int yoffs, xoffs, ysize, xsize;
  subband = steps;
  channel = LL;
  for (int i = 1; i <= steps && subband == steps && channel == LL; i++)
    {
      for (int j = 1; j <= 3 && subband == steps && channel == LL; j++)
        {
          where (areas[j], i, yoffs, xoffs, ysize, xsize);
          if (ypos >= yoffs && ypos < yoffs + ysize
              && xpos >= xoffs && xpos < xoffs + xsize)
            {
              subband = i;
              channel = areas[j];
            }
        }
    }
}

coeff 
WaveletTransform::ratio (int subband1, int subband2)
{
  return pow (4.0, (double)(subband2 - subband1));
}

int 
WaveletTransform::getSubband (int ypos, int xpos, int steps)
{
  area channel;
  int subband;
  where (ypos, xpos, subband, channel, steps);
  return subband;
}

area
WaveletTransform::getArea (int ypos, int xpos, int steps)
{
  area channel;
  int subband;
  where (ypos, xpos, subband, channel, steps);
  return channel;
}
